@@ -10,11 +10,7 @@ module Test.QuickCheck.GenericArbitrary (
 import GHC.Generics
 import Test.QuickCheck
 
-#if !MIN_VERSION_base(4,8,0)
-import Control.Applicative (pure, (<$>), (<*>))
-#endif
-
--- Generic arbitary for non-recursive types
+-- Generic arbitrary for non-recursive types
 genericArbitrary :: (Generic a, GArbitrary (Rep a)) => Gen a
 genericArbitrary = fmap to garbitrary
 
