@@ -13,8 +13,6 @@ import Distribution.Types.VersionRange (VersionRange)
 import Distribution.Types.GenericPackageDescription (GenericPackageDescription)
 import Distribution.Types.LocalBuildInfo            (LocalBuildInfo)
 
-import UnitTests.Orphans ()
-
 tests :: TestTree
 tests = testGroup "Distribution.Utils.Structured"
     -- This test also verifies that structureHash doesn't loop.
@@ -32,16 +30,8 @@ md5Check proxy md5Int = structureHash proxy @?= md5FromInteger md5Int
 
 md5CheckGenericPackageDescription :: Proxy GenericPackageDescription -> Assertion
 md5CheckGenericPackageDescription proxy = md5Check proxy
-#if MIN_VERSION_base(4,19,0)
-    0x7559521b9eb2e2fa4a608a86c629dc17
-#else
-    0xa78ea118e2e29b5809d359c9431df3ba
-#endif
+    0x09251b46ffc5178a7526d31e794d9c62
 
 md5CheckLocalBuildInfo :: Proxy LocalBuildInfo -> Assertion
 md5CheckLocalBuildInfo proxy = md5Check proxy
-#if MIN_VERSION_base(4,19,0)
-    0x8a8e81b52a34b8610acdcd0b9d488940
-#else
-    0xb53fbd58281a6f329f7b659d91fcd86e
-#endif
+    0x93b7e8ebb5b9f879fa5fe49b1708b43b
