@@ -3,8 +3,9 @@ import Test.Cabal.Prelude
 -- when linked dynamically
 -- See https://github.com/haskell/cabal/issues/4270
 main = setupAndCabalTest $ do
-  skipUnless "no shared libs"   =<< hasSharedLibraries
+  skipIfAllCabalVersion "< 2.2"
+  skipIfNoSharedLibraries
   skipUnless "no shared Cabal"  =<< hasCabalShared
-  skipUnless "no Cabal for GHC" =<< hasCabalForGhc
-  setup_build ["--enable-tests", "--enable-executable-dynamic"]
-  setup "test" []
+  expectBrokenIfOSXAndGhc "== 8.0.2" 8028 $ do
+    setup_build ["--enable-tests", "--enable-executable-dynamic"]
+    setup "test" []
