@@ -13,6 +13,13 @@ main _t = do
   copyFile (dir </> from) (dir </> to)
   T(expectEq) () [from, to] . List.sort =<< listDirectory dir
   T(expectEq) () contents =<< readFile (so (dir </> to))
+
+  -- Regression test for https://github.com/haskell/directory/issues/177
+  createDirectory "issue177"
+  T(expectIOErrorType) () isDoesNotExistError
+    (copyFile "issue177/nonexistentSrc" "issue177/dst")
+  T(expectEq) () [] =<< listDirectory "issue177"
+
   where
     contents = "This is the data\n"
     from     = "source"
